"""The pyTest Classes for testing the SlxJsonRpc Package."""
from enum import Enum

from typing import Any
from typing import List
# from typing import Literal  # NOTE: Not possible in py37
from typing import Union

import pytest

import slxjsonrpc


class MethodsTest(str, Enum):
    """The Enum of Methods for the SlXJsonRpc."""

    add = "add"
    ping = "ping"
    sub = "sub"
    crash = "crash"
    tweet = "tweet"
    error = "error"
    nop = 'NOP!'


class TestSlxJsonRpc:
    """Test the communication between the SlxJsonRpc Server & Client."""

    def setup_method(self):
        """Create the server & client instances for SlxJsonRpc."""
        self.tweet_data = None
        self.error_code = 0

        def tweeting(data):
            self.tweet_data = data

        def custom_error(*args, **kwargs):
            raise slxjsonrpc.RpcErrorException(
                code=self.error_code, msg="Just some Error!"
            )

        params = {
            MethodsTest.add: List[Union[int, float]],
            MethodsTest.sub: List[Union[int, float]],
            MethodsTest.ping: None,
            MethodsTest.crash: None,
            MethodsTest.tweet: Any,
            MethodsTest.error: Any,
            MethodsTest.nop: Any,
        }
        result = {
            MethodsTest.add: Union[int, float],
            MethodsTest.sub: Union[int, float],
            MethodsTest.ping: str,  # Literal["pong"]
            MethodsTest.crash: int,
            MethodsTest.tweet: None,
            MethodsTest.error: None,
            MethodsTest.nop: Any,
        }
        method_map = {
            MethodsTest.add: lambda data: sum(data),
            MethodsTest.sub: lambda data: data[0] - sum(data[1:]),
            MethodsTest.ping: lambda data: "pong",
            MethodsTest.crash: lambda *args: "*beep*" - 42,
            MethodsTest.tweet: lambda data: tweeting(data),
            MethodsTest.error: custom_error,
        }
        self.server = slxjsonrpc.SlxJsonRpc(
            methods=MethodsTest, result=result, params=params, method_cb=method_map
        )

    @pytest.mark.parametrize(
        "exclude_unset",
        [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_none", [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_defaults", [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "data_in,data_out",
        [
            [
                '{"jsonrpc":"2.0","method":"add","id":"s1","params": [1, 2, 3]}',
                '{"jsonrpc":"2.0","id":"s1","result":6}',
            ],
            [
                '{"jsonrpc":"2.0","method":"ping","id":"s1122"}',
                '{"jsonrpc":"2.0","id":"s1122","result":"pong"}',
            ],
            [
                '{"jsonrpc":"2.0","method":"ping"}',
                None,
            ],
            [
                '{"jsonrpc":"2.0","method":"tweet","params":"test"}',
                None,
            ],
            # # NOTE:Will fail until pydantic v2 have an option to force result.
            # [
            #     '{"jsonrpc":"2.0","method":"tweet","id":"hh","params":"test"}',
            #     '{"jsonrpc":"2.0","id":"hh","result":null}',
            # ],
        ],
    )
    def test_flow(
        self,
        data_in: str,
        data_out: str,
        exclude_unset: bool,
        exclude_none: bool,
        exclude_defaults: bool,
    ):
        """Testing the server Happy Flow."""
        model_data = self.server.parser(data_in)

        if data_out is None:
            assert model_data is None
            return
        str_data = model_data.model_dump_json(
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )
        assert str_data == data_out

    @pytest.mark.parametrize(
        "exclude_unset",
        [
            True,
            # False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_none", [
            True,
            # False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_defaults", [
            True,
            False
        ]
    )
    @pytest.mark.parametrize(
        "data_in,data_out",
        [
            [
                '{"jsonrpc": "2.0", "method"',
                (
                    '{"jsonrpc":"2.0","error":'
                    # '{"id":null,"jsonrpc":"2.0","error":' # NOTE: Right one!
                    '{"code":-32700,'
                    '"message":"Invalid JSON was received by the server."}}'
                ),
            ],
            [
                "",
                (
                    '{"jsonrpc":"2.0","error":'
                    # '{"id":null,"jsonrpc":"2.0","error":' # NOTE: Right one!
                    '{"code":-32700,'
                    '"message":"Invalid JSON was received by the server."}}'
                ),
            ],
            # # NOTE: Will fail until pydantic allow to force Exclude None.
            # [
            #     "[]",
            #     (
            #         '{"jsonrpc":"2.0","error":'
            #         '{"code":-32600,'
            #         '"message":"The JSON sent is not a valid Request object."}}'
            #     ),
            # ],
            [
                '{"foo":"boo"}',
                (
                    '{"jsonrpc":"2.0","error":'
                    # '{"id":null,"jsonrpc":"2.0","error":' # NOTE: Right one!
                    '{"code":-32600,'
                    '"message":"The JSON sent is not a valid Request object."}'
                    "}"
                ),
            ],
            [
                '{"jsonrpc":"2.0","method":"NOP!","params":"test", "id":"hej"}',
                (
                    '{"id":"hej","jsonrpc":"2.0",'
                    '"error":{"code":-32601,"message":"'
                    'The method does not exist / is not available."}}'
                ),
            ],
            [
                '{"jsonrpc":"2.0","method":"NOP!","params":"test"}',
                None,
            ],
            [
                '{"jsonrpc":"2.0","method":"NOWHERE!","id":"1q"}',
                (
                    '{"id":"1q","jsonrpc":"2.0","error":'
                    '{"code":-32601,'
                    '"message":"The method does not exist / is not available."}}'
                ),
            ],
            [
                '{"jsonrpc":"2.0","method":"NOWHERE!"}',
                None,
            ],
            [
                '{"jsonrpc":"2.0","method":"add","id":"-32s1","params":"NOP!"}',
                (
                    '{"id":"-32s1","jsonrpc":"2.0","error":'
                    '{"code":-32602,'
                    '"message":"Invalid method parameter(s)."}}'
                ),
            ],
            [
                '{"jsonrpc":"2.0","method":"add","params":"NOP!"}',
                None,
            ],
            [
                '{"jsonrpc":"2.0","method":"add","id":"s102"}',
                (
                    '{"id":"s102","jsonrpc":"2.0","error":'
                    '{"code":-32602,'
                    '"message":"Invalid method parameter(s)."}}'
                ),
            ],
            [
                '{"jsonrpc":"2.0","method":"add"}',
                None,
            ],
            [
                '{"jsonrpc":"2.0","method":"crash","id":"12342"}',
                (
                    '{"id":"12342","jsonrpc":"2.0","error":'
                    '{"code":-32000,'
                    '"message":"Internal server error."}}'
                ),
            ],
            [
                '{"jsonrpc":"2.0","method":"crash"}',
                None,
            ],
            # # [-32099, ''],
        ],
    )
    def test_request_with_errors(
        self,
        data_in: str,
        data_out: str,
        exclude_unset: bool,
        exclude_none: bool,
        exclude_defaults: bool,
    ):
        """Testing the Request Happy Flow."""
        model_data = self.server.parser(data_in)

        if data_out is None:
            assert model_data is None
            return
        str_data = model_data.model_dump_json(
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )
        assert str_data == data_out

    @pytest.mark.parametrize(
        "data_in,data_out",
        [
            [
                '{"id":"12342","jsonrpc":"2.0","error":{"code":-32000,"message":"Internal server error."}}',
                None,
            ],
            [
                '{"id":null,"jsonrpc":"2.0","error":{"code":-32000,"message":"Internal server error."}}',
                None,
            ],
        ],
    )
    def test_receive_of_error(
        self,
        data_in: str,
        data_out: str,
    ):
        """Testing if it do some unexpected when receiving an RpcError."""
        model_data = self.server.parser(data_in)

        assert model_data is None

    @pytest.mark.parametrize(
        "exclude_unset",
        [
            True,
            # False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_none",
        [
            # True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_defaults",
        [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "data_in,data_out",
        [
            [
                (
                    '[{"jsonrpc":"2.0","method":"add","id":"s1","params":[1,2,3]},'
                    '{"jsonrpc":"2.0","method":"ping","id":"s1122"}]'
                ),
                (
                    '[{"jsonrpc":"2.0","id":"s1","result":6},'
                    '{"jsonrpc":"2.0","id":"s1122","result":"pong"}]'
                ),
            ],
            [
                '[{"jsonrpc":"2.0","method":"add","id":"s1","params":[1,2,3]}]',
                '[{"jsonrpc":"2.0","id":"s1","result":6}]',
            ],
            [
                '[{"jsonrpc":"2.0","method":"add","params":[1,2,3]}]',
                None,
            ],
            [
                (
                    '[{"foo":"boo"},'
                    '{"jsonrpc":"2.0","method":"add","id":"s1","params":[1,2,3]}]'
                ),
                (
                    '[{"id":null,"jsonrpc":"2.0","error":'
                    '{"code":-32600,'
                    '"message":"The JSON sent is not a valid Request object."}},'
                    '{"jsonrpc":"2.0","id":"s1","result":6}]'
                ),
            ],
            [
                (
                    '[{"jsonrpc":"2.0","id":"ff","foo":"boo"},'
                    '{"jsonrpc":"2.0","method":"add","id":"s1","params":[1,2,3]}]'
                ),
                (
                    '[{"id":"ff","jsonrpc":"2.0","error":'
                    '{"code":-32600,'
                    '"message":"The JSON sent is not a valid Request object."}},'
                    '{"jsonrpc":"2.0","id":"s1","result":6}]'
                ),
            ],
        ],
    )
    def test_received_bulk(
        self,
        data_in: str,
        data_out: str,
        exclude_unset: bool,
        exclude_none: bool,
        exclude_defaults: bool,
    ):
        """Test if the Bulking receiving works as intended."""
        model_data = self.server.parser(data_in)

        if data_out is None:
            assert model_data is None
            return
        str_data = model_data.model_dump_json(
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )
        assert str_data == data_out

    @pytest.mark.skip("TBW: Make server only.")
    @pytest.mark.parametrize(
        "error_code",
        # list(range(-32099, -32000 + 1)),
        [-32099, -32050, -32000],
    )
    def test_custom_error_response(self, error_code):
        """Test if the custom error response works as intended."""
        self.error_code = error_code
        msg = '{"jsonrpc": "2.0", "method": "error", "id": "12342"}'
        error_obj = self.server.parser(msg)
        obj_code = (
            error_obj.error.code
            if isinstance(error_obj.error.code, int)
            else error_obj.error.code.value
        )
        assert obj_code == error_code

    @pytest.mark.parametrize(
        "data_in,data_out",
        [
            [
                '{"jsonrpc":"2.0","method":"add","id":"s1","params": [1, 2, 3]}',
                (
                    '{"id":"s1","jsonrpc":"2.0","error":'
                    '{"code":-32603,'
                    '"message":"Internal JSON-RPC error."}}'
                ),
            ],
            [
                '{"jsonrpc":"2.0","method":"add","params": [1, 2, 3]}',
                None,
            ],
        ],
    )
    def test_internal_error(
        self,
        data_in: str,
        data_out: str,
    ):
        """Testing the server Happy Flow."""
        backup = self.server._method_cb
        self.server._method_cb = None
        model_data = self.server.parser(data_in)

        if data_out is None:
            assert model_data is None
            return

        str_data = model_data.model_dump_json(
            exclude_none=True,
        )
        assert str_data == data_out

        self.server._method_cb = backup

    # @pytest.mark.parametrize(
    #     "data_in",
    #     [
    #         '{"jsonrpc":"2.1","id":"f","method":"ping"}',
    #     ],
    # )
    # def test_unknown_response_Type(
    #     self,
    #     data_in,
    # ):
    #     """For triggering the 'Unhanded ValidationError' logic."""
    #     #NOTE: Not possible.
    #     c_data = self.server.parser(
    #         data=data_in
    #     )
    #     print(c_data)
    #     assert False



"""The pyTest Classes for testing the SlxJsonRpc Package."""
from enum import Enum

from typing import List

# from typing import Literal  # NOTE: Not possible in py37
from typing import Union

import pytest

from slxjsonrpc.schema import jsonrpc as jsonrpc_schema
from slxjsonrpc.schema.jsonrpc import MethodError

from pydantic import ValidationError
from pydantic import BaseModel


class Point(BaseModel):
    """Coordinate Point Object."""
    x: int
    y: int


class MethodsTest(str, Enum):
    """The Enum of Methods for the SlXJsonRpc."""

    add = "add"
    ping = "ping"
    sub = "sub"
    point = "point"


class TestSchema:
    """Test the JsonRpc Schema."""

    def setup_method(self):
        """Setup the Schema mapping."""
        self.params_map = {
            MethodsTest.add: List[Union[int, float]],
            MethodsTest.sub: List[Union[int, float]],
            MethodsTest.ping: None,
            MethodsTest.point: Point,
        }
        jsonrpc_schema.set_params_map(
            self.params_map,
        )

    @pytest.mark.parametrize(
        "method,data,should_trigger_exception",
        [
            ["add", [1, 2, 3], False],
            ["add", "pong", True],
            ["sub", [1, 2, 3], False],
            ["sub", "pong", True],
            ["ping", None, False],
            ["ping", [1, 2, 3], True],
            ["NOP", None, True],
            ["NOP", "Nop!", True],
            ["point", Point(x=1, y=2), False],
        ],
    )
    def test_request(self, method, data, should_trigger_exception):
        """Test basic Request parsing."""
        jsonrpc_schema.RpcRequest.update_method(MethodsTest)
        strJson = {
            "id": "1",
            "jsonrpc": "2.0",
            "method": method,
        }
        if data:
            strJson["params"] = data
        print(strJson)

        try:
            r_data = jsonrpc_schema.RpcRequest.model_validate(strJson)
        except (ValidationError, MethodError):
            if not should_trigger_exception:
                raise
        else:
            if should_trigger_exception:
                raise ValueError(f"Should Not have passed: {r_data}")

    @pytest.mark.parametrize(
        "method,the_id,result,data_out",
        [
            [
                MethodsTest.point,
                'hey_id',
                Point(x=1, y=2),
                '{"jsonrpc":"2.0","id":"hey_id","result":{"x":1,"y":2}}'
            ],
            [
                'something',
                'hey_id1',
                None,
                '{"jsonrpc":"2.0","id":"hey_id1","result":null}'
            ],
        ]
    )
    def test_Response(self, method, the_id, result, data_out):
        """."""
        jsonrpc_schema.set_id_mapping({the_id: method})
        jsonrpc_schema.set_result_map({method: type(result)})
        try:
            r_data = jsonrpc_schema.RpcResponse(
                jsonrpc=jsonrpc_schema.RpcVersion.v2_0,
                id=the_id,
                result=result,
            )

            assert data_out == r_data.model_dump_json()
        finally:
            jsonrpc_schema.set_id_mapping({})
            jsonrpc_schema.set_result_map({})

    @pytest.mark.parametrize(
        "method,data,should_trigger_exception",
        [
            ["add", [1, 2, 3], False],
            ["add", "pong", True],
            ["sub", [1, 2, 3], False],
            ["sub", "pong", True],
            ["ping", None, False],
            ["ping", [1, 2, 3], True],
            ["NOP", None, True],
            ["NOP", "Nop!", True],
            ["point", Point(x=1, y=2), False],
        ],
    )
    def test_notifications(self, method, data, should_trigger_exception):
        """Test basic Notification parsing."""
        jsonrpc_schema.RpcNotification.update_method(MethodsTest)
        strJson = {
            "jsonrpc": "2.0",
            "method": method,
        }
        if data:
            strJson["params"] = data
        print(strJson)

        try:
            r_data = jsonrpc_schema.RpcNotification.model_validate(strJson)
        except (ValidationError, MethodError):
            if not should_trigger_exception:
                raise
        else:
            if should_trigger_exception:
                raise ValueError(f"Should Not have passed: {r_data}")

    @pytest.mark.parametrize(
        "_id",
        [
            "hej",
            "øæasdnh",
            "he123\"}kjnf",
        ]
    )
    def test_response_id(self, _id):
        """Test if the variable naming convention works."""
        method = "add"
        data = [1, 2, 3]
        jsonrpc_schema.RpcRequest.update_method(MethodsTest)
        jsonrpc_schema.rpc_set_name(...)
        strJson = {
            "id": _id,
            "jsonrpc": "2.0",
            "method": method,
        }
        if data:
            strJson["params"] = data

        r_data = jsonrpc_schema.RpcRequest.model_validate(strJson)

        assert r_data.id == _id

    @pytest.mark.parametrize(
        "_id",
        [
            "hej",
            "øæasdnh",
            "he123\"}kjnf",
        ]
    )
    def test_emitted_id(self, _id):
        """Test if the variable naming convention works."""
        method = "add"
        data = [1, 2, 3]
        jsonrpc_schema.RpcRequest.update_method(MethodsTest)
        jsonrpc_schema.rpc_set_name(_id)

        r_data = jsonrpc_schema.RpcRequest(
            method=method,
            params=data,
        )

        assert r_data.id == f"{jsonrpc_schema._session_id}_{_id}_{jsonrpc_schema._session_count}"

    @pytest.mark.parametrize(
        "_id",
        [
            "hej",
            "øæasdnh",
            "he123\"}kjnf",
        ]
    )
    def test_given_id(self, _id):
        """Test if the variable naming convention works."""
        method = "add"
        data = [1, 2, 3]
        jsonrpc_schema.RpcRequest.update_method(MethodsTest)
        jsonrpc_schema.rpc_set_name(...)

        r_data = jsonrpc_schema.RpcRequest(
            method=method,
            params=data,
            id=_id,
        )

        assert r_data.id == _id

    @pytest.mark.parametrize(
        "data_in,the_id,method,params",
        [
            (
                '{"jsonrpc":"2.0","method":"point","id":"some_id","params":{"x":1, "y": 2}}',
                'some_id',
                'point',
                Point(x=1, y=2),
            )
        ],
    )
    def test_request_deserializing(self, data_in, the_id, method, params):
        """Test deserializing of a json object."""
        data = jsonrpc_schema.RpcRequest.model_validate_json(data_in)

        assert data.method == method
        assert data.id == the_id
        assert data.params == params

    @pytest.mark.parametrize(
        "data_in,method,params",
        [
            (
                '{"jsonrpc":"2.0","method":"point","params":{"x":1, "y": 2}}',
                'point',
                Point(x=1, y=2),
            )
        ],
    )
    def test_notification_deserializing(self, data_in, method, params):
        """Test deserializing of a json object."""
        data = jsonrpc_schema.RpcNotification.model_validate_json(data_in)

        assert data.method == method
        assert data.params == params

    @pytest.mark.parametrize(
        "method,params,data_out",
        [
            [
                'method',
                'data',
                '{{"jsonrpc":"2.0","method":"method","id":"{}","params":"data"}}'
            ],
        ]
    )
    def test_no_mapping_Request(self, method, params, data_out):
        """For testing the Request if no params where set."""
        jsonrpc_schema.set_params_map({})
        try:
            r_data = jsonrpc_schema.RpcRequest(
                jsonrpc=jsonrpc_schema.RpcVersion.v2_0,
                method=method,
                params=params,
            )
            the_id = r_data.id
            data_string = data_out.format(the_id)
            assert data_string == r_data.model_dump_json()
        finally:
            jsonrpc_schema.set_params_map(self.params_map)

    @pytest.mark.parametrize(
        "method,params,data_out",
        [
            [
                'method',
                'data',
                '{"jsonrpc":"2.0","method":"method","params":"data"}'
            ],
        ]
    )
    def test_no_mapping_notification(self, method, params, data_out):
        """For test Notification if no method to params mapping where set."""
        jsonrpc_schema.set_params_map({})
        try:
            r_data = jsonrpc_schema.RpcNotification(
                jsonrpc=jsonrpc_schema.RpcVersion.v2_0,
                method=method,
                params=params,
            )

            assert data_out == r_data.model_dump_json()
        finally:
            jsonrpc_schema.set_params_map(self.params_map)

    @pytest.mark.parametrize(
        "the_id,result,data_out",
        [
            [
                'hey_id',
                Point(x=1, y=2),
                '{"jsonrpc":"2.0","id":"hey_id","result":{"x":1,"y":2}}'
            ],
        ]
    )
    def test_no_mapping_Response(self, the_id, result, data_out):
        """."""
        jsonrpc_schema.set_id_mapping({})
        jsonrpc_schema.set_result_map({})

        r_data = jsonrpc_schema.RpcResponse(
            jsonrpc=jsonrpc_schema.RpcVersion.v2_0,
            id=the_id,
            result=result,
        )

        assert data_out == r_data.model_dump_json()

    @pytest.mark.parametrize(
        "the_id,result,data_out",
        [
            [
                'hey_id',
                Point(x=1, y=2),
                '{"jsonrpc":"2.0","id":"hey_id","result":{"x":1,"y":2}}'
            ],
        ]
    )
    def test_error_config_Response(self, the_id, result, data_out):
        """."""
        jsonrpc_schema.set_id_mapping({the_id: 'FAKE_METHOD'})
        jsonrpc_schema.set_result_map({'FAKE_METHOD': None})

        try:
            with pytest.raises(ValueError):
                jsonrpc_schema.RpcResponse(
                    jsonrpc=jsonrpc_schema.RpcVersion.v2_0,
                    id=the_id,
                    result=result,
                )
        finally:
            jsonrpc_schema.set_id_mapping({})
            jsonrpc_schema.set_result_map({})

    @pytest.mark.parametrize(
        "the_id,result,data_out",
        [
            [
                'hey_id',
                Point(x=1, y=2),
                '{"jsonrpc":"2.0","id":"hey_id","result":{"x":1,"y":2}}'
            ],
        ]
    )
    def test_missing_mapping_Response(self, the_id, result, data_out):
        """."""
        jsonrpc_schema.set_id_mapping({the_id: 'FAKE_METHOD'})
        jsonrpc_schema.set_result_map({MethodsTest.point: Point})

        try:
            with pytest.raises(ValueError):
                jsonrpc_schema.RpcResponse(
                    jsonrpc=jsonrpc_schema.RpcVersion.v2_0,
                    id=the_id,
                    result=result,
                )
        finally:
            jsonrpc_schema.set_id_mapping({})
            jsonrpc_schema.set_result_map({})



"""The pyTest Classes for testing the SlxJsonRpc Package."""
from enum import Enum

from typing import Any
from typing import List

# from typing import Literal  # NOTE: Not possible in py37
from typing import Union

import pytest

import slxjsonrpc


# slxjsonrpc.schema.jsonrpc._id_gen = lambda *args, **kwargs: 'the_id'

class MethodsTest(str, Enum):
    """The Enum of Methods for the SlXJsonRpc."""

    add = "add"
    ping = "ping"
    sub = "sub"
    crash = "crash"
    tweet = "tweet"
    error = "error"


class TestSlxJsonRpc:
    """Test the communication between the SlxJsonRpc Server & Client."""

    def setup_method(self):
        """Setup the server & client instances of SlxJsonRpc."""
        params = {
            MethodsTest.add: List[Union[int, float]],
            MethodsTest.sub: List[Union[int, float]],
            MethodsTest.ping: None,
            MethodsTest.crash: None,
            MethodsTest.tweet: Any,
            MethodsTest.error: Any,
        }
        result = {
            MethodsTest.add: Union[int, float],
            MethodsTest.sub: Union[int, float],
            MethodsTest.ping: str,  # Literal["pong"]
            MethodsTest.crash: int,
            MethodsTest.tweet: None,
            MethodsTest.error: None,
        }

        self.client = slxjsonrpc.SlxJsonRpc(
            methods=MethodsTest,
            result=result,
            params=params,
        )

    @pytest.mark.parametrize(
        "exclude_unset",
        [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_none", [
            True,
            # False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_defaults", [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "method,params,data_out",
        [
            [
                "ping",
                None,
                '{"jsonrpc":"2.0","method":"ping","id":"the_id"}',
            ],
            [
                "add",
                [1, 2, 3],
                '{"jsonrpc":"2.0","method":"add","id":"the_id","params":[1,2,3]}',
            ],
            [
                "sub",
                [1, 2, 3],
                '{"jsonrpc":"2.0","method":"sub","id":"the_id","params":[1,2,3]}'
            ],
        ],
    )
    def test_request(
        self,
        method,
        params,
        data_out,
        exclude_unset,
        exclude_none,
        exclude_defaults,
    ):
        """Testing the Request Happy Flow."""
        c_data = self.client.create_request(
            method=method,
            params=params,
            callback=lambda data: None
        )

        assert c_data is not None

        c_data.id = 'the_id'

        r_data = c_data.model_dump_json(
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )

        assert r_data == data_out

    @pytest.mark.parametrize(
        "data_in",
        [
            '{"jsonrpc":"2.0","id":"NeverUsedId","result":"pong"}',
        ],
    )
    def test_unknown_response(
        self,
        data_in,
    ):
        """For triggering the unknown ID in the response logic."""
        c_data = self.client.parser(
            data=data_in
        )
        assert c_data is None

    @pytest.mark.parametrize(
        "exclude_unset",
        [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_none", [
            True,
            # False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_defaults", [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "method,params,data_out",
        [
            [
                "ping",
                None,
                '{"jsonrpc":"2.0","method":"ping"}',
            ],
            [
                "add",
                [1, 2, 3],
                '{"jsonrpc":"2.0","method":"add","params":[1,2,3]}',
            ],
            [
                "sub",
                [1, 2, 3],
                '{"jsonrpc":"2.0","method":"sub","params":[1,2,3]}'
            ],
        ],
    )
    def test_notification(
        self,
        method,
        params,
        data_out,
        exclude_unset,
        exclude_none,
        exclude_defaults,
    ):
        """Testing the notification Happy Flow."""
        c_data = self.client.create_notification(
            method=method,
            params=params,
        )

        assert c_data is not None

        r_data = c_data.model_dump_json(
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )

        assert r_data == data_out

    @pytest.mark.parametrize(
        "exclude_unset",
        [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_none", [
            True,
            # False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_defaults", [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "method_params,data_out",
        [
            [
                [("ping", None), ("add", [1, 2, 3]), ("sub", [1, 2, 3])],
                (
                    '[{"jsonrpc":"2.0","method":"ping"},'
                    '{"jsonrpc":"2.0","method":"add","params":[1,2,3]},'
                    '{"jsonrpc":"2.0","method":"sub","params":[1,2,3]}]'
                ),
            ],
            [
                [("ping", None),],
                (
                    '{"jsonrpc":"2.0","method":"ping"}'
                ),
            ],
            [
                [],
                None,
            ],
        ],
    )
    def test_bulk(
        self,
        method_params,
        data_out,
        exclude_unset,
        exclude_none,
        exclude_defaults,
    ):
        """Test is the Bulking works as intended."""
        with self.client.batch():
            for method, params in method_params:
                c_data = self.client.create_notification(
                    method=method,
                    params=params,
                )
                assert c_data is None

        data = self.client.get_batch_data()

        if data_out is None:
            assert data is None
            return

        if isinstance(data, slxjsonrpc.RpcBatch):
            assert data[0]
            count = 0
            for x in data:
                count += 1
            assert count == len(data)

        r_data = data.model_dump_json(
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )

        assert r_data == data_out

    @pytest.mark.parametrize(
        "exclude_unset",
        [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_none", [
            True,
            # False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_defaults", [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "method_params,mp_thread,data_out",
        [
            [
                [("ping", None), ("add", [1, 2, 3])],
                ("sub", [1, 2, 3]),
                (
                    '[{"jsonrpc":"2.0","method":"ping"},'
                    '{"jsonrpc":"2.0","method":"add","params":[1,2,3]},'
                    '{"jsonrpc":"2.0","method":"sub","params":[1,2,3]}]'
                ),
            ],
        ],
    )
    def test_bulk_treaded(
        self,
        method_params,
        mp_thread,
        data_out,
        exclude_unset,
        exclude_none,
        exclude_defaults,
    ):
        """Test is the Bulking works as intended."""
        with self.client.batch():
            for method, params in method_params:
                c_data = self.client.create_notification(
                    method=method,
                    params=params,
                )
                assert c_data is None

        t_data = self.client.create_notification(
            method=mp_thread[0],
            params=mp_thread[1],
        )

        data = self.client.get_batch_data(t_data)

        if data_out is None:
            assert data is None
            return

        if isinstance(data, slxjsonrpc.RpcBatch):
            assert data[0]
            count = 0
            for x in data:
                count += 1
            assert count == len(data)

        r_data = data.model_dump_json(
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )

        assert r_data == data_out

    # @pytest.mark.parametrize(  # NOTE: Breaks because of the ig_gen hack
    #     "exclude_unset",
    #     [
    #         True,
    #         False,
    #     ],
    # )
    @pytest.mark.parametrize(
        "exclude_none", [
            True,
            # False,
        ],
    )
    @pytest.mark.parametrize(
        "exclude_defaults", [
            True,
            False,
        ],
    )
    @pytest.mark.parametrize(
        "method,params,data_out,data_in,code",
        [
            [
                "crash",
                None,
                '{"jsonrpc":"2.0","method":"crash","id":"the_id"}',
                (
                    '{"id":"the_id","jsonrpc":"2.0","error":'
                    '{"code":-32000,'
                    '"message":"Internal server error."}}'
                ),
                -32000,
            ],
            [
                "crash",
                None,
                '{"jsonrpc":"2.0","method":"crash","id":"the_id"}',
                (
                    '{"id":"the_id","jsonrpc":"2.0","error":'
                    '{"code":-32000,'
                    '"message":"Internal server error."}}'
                ),
                None,
            ],
        ],
    )
    def test_request_error_callback(
        self,
        method,
        params,
        data_out,
        data_in,
        # exclude_unset,  # NOTE: Breaks because of the ig_gen hack
        exclude_none,
        exclude_defaults,
        code,
    ):
        """."""
        backup = slxjsonrpc.schema.jsonrpc._id_gen
        slxjsonrpc.schema.jsonrpc._id_gen = lambda *args, **kwargs: 'the_id'
        error = None

        def err_cb(error_model) -> None:
            nonlocal error
            error = error_model

        c_data = self.client.create_request(
            method=method,
            params=params,
            callback=lambda data: None,
            error_callback=err_cb if code is not None else None,
        )

        slxjsonrpc.schema.jsonrpc._id_gen = backup

        assert c_data is not None

        print(c_data)

        r_data = c_data.model_dump_json(
            # exclude_unset=exclude_unset,  # NOTE: Breaks because of the ig_gen hack
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )

        assert r_data == data_out

        data = self.client.parser(data_in)

        assert data is None

        if code is None:
            return

        assert error is not None
        assert error.code == code

    def test_callback_example(self):
        """Just to get the code Coverage for the function example."""
        slxjsonrpc.jsonrpc.method_callback_example(1)

    @pytest.mark.parametrize(
        "method,params",
        [
            [
                "ping",
                'Shoudl\'tBeHere',
            ],
        ],
    )
    def test_misconfigured_request(
        self,
        method,
        params,
    ):
        """Testing the Request Happy Flow."""
        with pytest.raises(ValueError):
            self.client.create_request(
                method=method,
                params=params,
                callback=lambda data: None
            )

    @pytest.mark.parametrize(
        "method,params",
        [
            [
                "ping",
                'Shoudl\'tBeHere',
            ],
        ],
    )
    def test_misconfigured_notification(
        self,
        method,
        params,
    ):
        """Testing the Request Happy Flow."""
        with pytest.raises(ValueError):
            self.client.create_notification(
                method=method,
                params=params,
            )



"""The pyTest Classes for testing the SlxJsonRpc Package."""
from enum import Enum

from typing import Any
from typing import List

# from typing import Literal  # NOTE: Not possible in py37
from typing import Union

import pytest

import slxjsonrpc

from pydantic import BaseModel


class Point(BaseModel):
    """Coordinate Point Object."""
    x: int
    y: int


class MethodsTest(str, Enum):
    """The Enum of Methods for the SlXJsonRpc."""

    add = "add"
    ping = "ping"
    sub = "sub"
    crash = "crash"
    tweet = "tweet"
    error = "error"
    point = "point"


class TestSlxJsonRpc:
    """Test the communication between the SlxJsonRpc Server & Client."""

    def setup_method(self):
        """Setup the server & client instances of SlxJsonRpc."""
        self.tweet_data = None
        self.error_code = 0

        def tweeting(data):
            self.tweet_data = data

        def custom_error(*args, **kwargs):
            raise slxjsonrpc.RpcErrorException(
                code=self.error_code,
                msg="Just some Error!",
            )

        params = {
            MethodsTest.add: List[Union[int, float]],
            MethodsTest.sub: List[Union[int, float]],
            MethodsTest.ping: None,
            MethodsTest.crash: None,
            MethodsTest.tweet: Any,
            MethodsTest.error: Any,
            MethodsTest.point: Point,
        }
        result = {
            MethodsTest.add: Union[int, float],
            MethodsTest.sub: Union[int, float],
            MethodsTest.ping: str,  # Literal["pong"]
            MethodsTest.crash: int,
            MethodsTest.tweet: None,
            MethodsTest.error: None,
            MethodsTest.point: Point,
        }
        method_map = {
            MethodsTest.add: lambda data: sum(data),
            MethodsTest.sub: lambda data: data[0] - sum(data[1:]),
            MethodsTest.ping: lambda data: "pong",
            MethodsTest.crash: lambda *args: "*beep*" - 42,
            MethodsTest.tweet: lambda data: tweeting(data),
            MethodsTest.error: custom_error,
            MethodsTest.point: lambda data: Point(x=data.x * 2, y=data.y * 2),
        }
        self.server = slxjsonrpc.SlxJsonRpc(
            methods=MethodsTest,
            result=result,
            params=params,
            method_cb=method_map
        )
        self.client = slxjsonrpc.SlxJsonRpc(
            methods=MethodsTest,
            result=result,
            params=params,
            method_cb=method_map
        )

    @pytest.mark.parametrize(
        "method,params,result",
        [
            ["ping", None, "pong"],
            ["add", [1, 2, 3], 6],
            ["sub", [1, 2, 3], -4],
            ['point', Point(x=1, y=2), Point(x=2, y=4)],
        ],
    )
    def test_request_happy_flow(self, method, params, result):
        """Testing the Request Happy Flow."""
        round_trip = None

        def set_data(temp):
            nonlocal round_trip
            round_trip = temp

        c_data = self.client.create_request(
            method=method,
            params=params,
            callback=set_data
        )

        s_data = self.server.parser(c_data.model_dump_json(exclude_none=True))
        self.client.parser(s_data.model_dump_json(exclude_none=True))

        assert round_trip == result

    @pytest.mark.parametrize(
        "method,params,result",
        [
            ["tweet", "Trumphy", "Trumphy"],
            ["tweet", 1, 1],
            ['point', Point(x=1, y=2), None],
        ],
    )
    def test_notification_happy_flow(self, method, params, result):
        """Testing the Request Happy Flow."""
        c_data = self.client.create_notification(
            method=method,
            params=params
        )

        s_data = self.server.parser(c_data.model_dump_json(exclude_none=True))

        assert s_data is None
        if result:
            assert self.tweet_data == result

    @pytest.mark.parametrize(
        "error_code,data",
        [
            [-32700, '{"jsonrpc": "2.0", "method"'],
            [-32700, ""],
            [-32600, "[]"],
            [-32600, '{"foo": "boo"}'],
            [-32601, '{"jsonrpc": "2.0", "method": "NOWHERE!", "id": "1q"}'],
            [None, '{"jsonrpc": "2.0", "method": "NOWHERE!"}'],
            [-32602, '{"jsonrpc": "2.0", "method": "add", "id": "s1", "params": "NOP!"}'],
            [None, '{"jsonrpc": "2.0", "method": "add", "params": "NOP!"}'],
            [-32602, '{"jsonrpc": "2.0", "method": "add", "id": "s1"}'],
            [None, '{"jsonrpc": "2.0", "method": "add"}'],
            [-32000, '{"jsonrpc": "2.0", "method": "crash", "id": "12342"}'],
            [None, '{"jsonrpc": "2.0", "method": "crash"}'],
            # [-32099, ''],
        ],
    )
    def test_request_errors(self, data, error_code):
        """Testing the Request Happy Flow."""
        s_data = self.server.parser(data)
        print(f"{s_data}")
        if error_code is None:
            assert s_data is None
        else:
            assert s_data.error.code.value == error_code

    @pytest.mark.skip(reason="TBW!")
    @pytest.mark.parametrize("error_code, transformer", [(1, 2)])
    def test_return_types(self, error_code, transformer):
        """Testing if the return type is the right one."""
        error_obj = None

        def set_data(temp):
            nonlocal error_obj
            error_obj = temp

        c_data = self.client.create_request()
        s_data = self.server.parser(c_data.model_dump_json(exclude_none=True))
        e_data = transformer(s_data)
        r_data = self.client.parser(e_data)

        assert r_data is None

        assert error_obj.code.value == error_code

    @pytest.mark.parametrize(
        "method,params",
        [
            [MethodsTest.ping, None],
            [MethodsTest.add, [1, 2, 3]],
            [MethodsTest.sub, [1, 2, 3]],
        ],
    )
    @pytest.mark.parametrize(
        "error_code, transformer",
        [
            [
                -32700,
                lambda data: {
                    "jsonrpc": "2.0",
                    "id": data.id,
                    "error": {"code": -32700, "message": "", "data": "k"},
                },
            ],
            [
                -32600,
                lambda data: {
                    "jsonrpc": "2.0",
                    "id": data.id,
                    "error": {"code": -32600, "message": "", "data": "k"},
                },
            ],
            [
                -32601,
                lambda data: {
                    "jsonrpc": "2.0",
                    "id": data.id,
                    "error": {"code": -32601, "message": "", "data": "k"},
                },
            ],
            [
                -32602,
                lambda data: {
                    "jsonrpc": "2.0",
                    "id": data.id,
                    "error": {"code": -32602, "message": "", "data": "k"},
                },
            ],
            [
                -32603,
                lambda data: {
                    "jsonrpc": "2.0",
                    "id": data.id,
                    "error": {"code": -32603, "message": "", "data": "k"},
                },
            ],
            [
                -32000,
                lambda data: {
                    "jsonrpc": "2.0",
                    "id": data.id,
                    "error": {"code": -32000, "message": "", "data": "k"},
                },
            ],
        ],
    )
    def test_error_response(self, method, params, error_code, transformer):
        """Testing handling of the response, when receiving an RpcError."""
        error_obj = None
        data_obj = None

        def set_error(temp):
            nonlocal error_obj
            error_obj = temp

        def set_data(temp):
            nonlocal data_obj
            data_obj = temp

        c_data = self.client.create_request(
            method=method,
            params=params,
            callback=set_data,
            error_callback=set_error
        )
        s_data = self.server.parser(c_data.model_dump_json(exclude_none=True))
        e_data = transformer(s_data)
        r_data = self.client.parser(e_data)

        print(f"{r_data}")

        assert r_data is None
        assert data_obj is None
        assert error_obj.code.value == error_code

    def test_send_bulk(self):
        """Test is the Bulking works as intended."""
        cb_data = None

        def cb(data):
            print(f"data: {data}")
            nonlocal cb_data
            cb_data = data

        with self.client.batch():
            s_data = self.client.create_request(
                method=MethodsTest.add, callback=cb, params=[1, 2, 3]
            )
            assert s_data is None

            s_data = self.client.create_notification(method=MethodsTest.ping)
            assert s_data is None

        assert self.client.batch_size() == 2

        c_data = self.client.get_batch_data()
        # print(f"{s_data.model_dump_json(exclude_none=True)}")
        s_data = self.server.parser(c_data.model_dump_json(exclude_none=True))
        # print(data)
        assert len(s_data) == 1

        c_l_data = self.client.parser(s_data.model_dump_json(exclude_none=True))
        assert c_l_data is None
        assert cb_data == 6

    def test_received_bulk(self):
        """Test if the Bulking receiving works as intended."""
        pass

    @pytest.mark.parametrize(
        "error_code",
        # list(range(-32099, -32000 + 1)),
        [-32099, -32050, -32000],
    )
    def test_custom_error_response(self, error_code):
        """Test if the custom error response works as intended."""
        self.error_code = error_code
        msg = '{"jsonrpc": "2.0", "method": "error", "id": "12342"}'
        error_obj = self.client.parser(msg)
        obj_code = (
            error_obj.error.code
            if isinstance(error_obj.error.code, int)
            else error_obj.error.code.value
        )
        assert obj_code == error_code

    def test_unknown_id(self):
        """Test if the received jsonRps id is unknown."""
        pass
